@@ -1,33 +1,29 @@
-# This file is part of wildboar
-#
-# wildboar is free software: you can redistribute it and/or modify it
-# under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# wildboar is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
 # Authors: Isak Samsten
+# License: BSD 3 clause
 
+import base64
 import hashlib
-import io
 import os
+import pickle
 import re
 import sys
+import warnings
 import zipfile
-from abc import abstractmethod, ABCMeta
+from abc import ABCMeta, abstractmethod
 
 import numpy as np
-import requests
-from wildboar import __version__ as wildboar_version
 from pkg_resources import parse_version
-from scipy.io.arff import loadarff
+
+from .. import __version__ as wildboar_version
+from .. import eos as EOS
+
+try:
+    import requests
+except ModuleNotFoundError as e:
+    from ..utils import _soft_dependency_error
+
+    _soft_dependency_error(e, context="wildboar.datasets")
+
 
 DEFAULT_TAG = "default"
 
@@ -52,7 +48,7 @@ def _replace_placeholders(url, **kwargs):
     return url
 
 
-def _check_integrity(bundle_file, hash_file):
+def _check_integrity(bundle_file, hash_file, throws=True):
     """Check the integrity of the downloaded or cached file
 
     Parameters
@@ -63,6 +59,9 @@ def _check_integrity(bundle_file, hash_file):
     hash_file : str, bytes or PathLike
         Path to the hash file
 
+    throws : bool
+        Throw an exception on hash missmatch
+
     Returns
     -------
     bool : true if the hash of bundle file matches the contents of the hash file
@@ -73,9 +72,13 @@ def _check_integrity(bundle_file, hash_file):
     if hash is not None:
         actual_hash = _sha1(bundle_file)
         if hash != actual_hash:
-            raise ValueError(
-                "integrity check failed, expected '%s', got '%s'" % (hash, actual_hash)
-            )
+            if throws:
+                raise ValueError(
+                    "Integrity check failed, expected '%s', got '%s'."
+                    % (hash, actual_hash)
+                )
+            else:
+                return False
     return True
 
 
@@ -136,7 +139,8 @@ def _load_archive(
             os.makedirs(os.path.abspath(cache_dir), exist_ok=True)
         else:
             raise ValueError(
-                "output directory does not exist (set create_cache_dir=True to create it)"
+                "The output directory %s does not exist. "
+                "Set create_cache_dir=True to create it." % cache_dir
             )
 
     cached_hash = os.path.join(cache_dir, "%s.sha1" % bundle_name)
@@ -151,7 +155,9 @@ def _load_archive(
     if not os.path.exists(cached_hash) or not _sha1_is_sane(cached_hash):
         _download_hash_file(cached_hash, "%s.sha1" % download_url, bundle_name)
 
-    if os.path.exists(cached_bundle) and _check_integrity(cached_bundle, cached_hash):
+    if os.path.exists(cached_bundle) and _check_integrity(
+        cached_bundle, cached_hash, throws=False
+    ):
         try:
             return zipfile.ZipFile(open(cached_bundle, "rb"))
         except zipfile.BadZipFile:
@@ -178,16 +184,64 @@ def _download_hash_file(cached_hash, hash_url, filename):
     """
     with open(cached_hash, "w") as f:
         response = requests.get(hash_url)
-        if not response:
+        try:
+            response.raise_for_status()
+        except requests.exceptions.HTTPError as e:
             f.close()
             os.remove(cached_hash)
             raise ValueError(
-                "bundle (%s) not found (.sha1-file is missing). Try another version or tag."
-                % filename
-            )
+                "Downloading %s.sha1 failed. Try another version or tag." % filename
+            ) from e
+
         f.write(response.text)
 
 
+_SIZE_NAMES = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
+
+
+def _convert_size(byte_size, i=None):
+    import math
+
+    if byte_size == 0:
+        return 0, 0
+
+    if i is None:
+        i = int(math.floor(math.log(byte_size, 1024)))
+    p = math.pow(1024, i)
+    s = round(byte_size / p, 2)
+    return s, i
+
+
+def _print_progress(
+    current_iter,
+    max_iter,
+    *,
+    prefix="",
+    suffix="",
+    progress=None,
+    length=50,
+    fill="█",
+    unfill="-",
+    end="\r",
+    file=sys.stderr,
+):
+    def size_progress(current, total):
+        total_size, total_type = _convert_size(total)
+        current_size, _ = _convert_size(current, total_type)
+        return f" {current_size}/{total_size} {_SIZE_NAMES[total_type]}"
+
+    if progress is None:
+        progress = size_progress
+
+    progress = progress(float(current_iter), float(max_iter))
+    fill_size = int(length * current_iter // max_iter)
+    unfill_size = int(length - fill_size)
+    bar = fill * fill_size + unfill * unfill_size
+    print(f"\r{prefix}|{bar}|{progress}{suffix}", file=file, end=end)
+    if current_iter >= max_iter:
+        print(file=file)
+
+
 def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
     """Download the bundle
 
@@ -207,13 +261,14 @@ def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
     """
     with open(cached_bundle, "wb") as f:
         response = requests.get(bundle_url, stream=True)
-        if not response:
+        try:
+            response.raise_for_status()
+        except requests.exceptions.HTTPError as e:
             f.close()
             os.remove(cached_bundle)
             raise ValueError(
-                "bundle (%s) not found (.zip-file is missing). Try another version or tag."
-                % filename
-            )
+                "Downloading %s.zip failed. Try another version or tag." % filename
+            ) from e
 
         total_length = response.headers.get("content-length")
         if total_length is None:  # no content length header
@@ -221,27 +276,46 @@ def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
         else:
             length = 0
             total_length = int(total_length)
+
+            if progress:
+                total_size, total_type = _convert_size(total_length)
+                print(
+                    f"Downloading {filename}.zip "
+                    f"({total_size} {_SIZE_NAMES[total_type]})",
+                    file=sys.stderr,
+                )
+
             for data in response.iter_content(chunk_size=4096):
                 length += len(data)
                 f.write(data)
-                done = int(50 * length / total_length)
-                if length % 10 == 0 and progress:
-                    sys.stderr.write(
-                        "\r[%s%s] %d/%d downloading %s"
-                        % (
-                            "=" * done,
-                            " " * (50 - done),
-                            length,
-                            total_length,
-                            filename,
-                        )
+                if progress:
+                    _print_progress(
+                        length,
+                        total_length,
+                        prefix="  ",
+                        unfill="-",
                     )
-                    sys.stderr.flush()
 
 
 class Repository(metaclass=ABCMeta):
     """A repository is a collection of bundles"""
 
+    def __init__(self):
+        self._active = False
+
+    def __eq__(self, o):
+        if isinstance(o, Repository):
+            return self.identifier == o.identifier
+        return False
+
+    def __hash__(self) -> int:
+        return hash(self.identifier)
+
+    @property
+    @abstractmethod
+    def identifier(self):
+        pass
+
     @property
     @abstractmethod
     def name(self):
@@ -296,6 +370,10 @@ class Repository(metaclass=ABCMeta):
         """
         pass
 
+    @property
+    def active(self):
+        return self._active
+
     def get_bundle(self, key):
         """Get a bundle with the specified key
 
@@ -311,7 +389,7 @@ class Repository(metaclass=ABCMeta):
         """
         bundle = self.get_bundles().get(key)
         if bundle is None:
-            raise ValueError("bundle (%s) does not exist" % key)
+            raise ValueError("The bundle %s does not exist." % key)
         return bundle
 
     def load_dataset(
@@ -324,14 +402,13 @@ class Repository(metaclass=ABCMeta):
         tag=None,
         create_cache_dir=True,
         progress=True,
-        dtype=None,
         force=False,
     ):
         bundle = self.get_bundle(bundle)
         version = version or bundle.version
-        tag = tag or DEFAULT_TAG
+        tag = tag or bundle.tag
 
-        cache_dir = os.path.join(cache_dir, self.name)
+        cache_dir = os.path.join(cache_dir, self.identifier)
         download_url = _replace_placeholders(
             self.download_url, bundle=bundle.key, version=version, tag=tag
         )
@@ -344,13 +421,14 @@ class Repository(metaclass=ABCMeta):
             progress=progress,
             force=force,
         ) as archive:
-            return bundle.load(dataset, archive, dtype=dtype)
+            return bundle.load(dataset, archive)
 
     def list_datasets(
         self,
         bundle,
         *,
         cache_dir,
+        collection=None,
         version=None,
         tag=None,
         create_cache_dir=True,
@@ -359,9 +437,9 @@ class Repository(metaclass=ABCMeta):
     ):
         bundle = self.get_bundle(bundle)
         version = version or bundle.version
-        tag = tag or DEFAULT_TAG
+        tag = tag or bundle.tag
 
-        cache_dir = os.path.join(cache_dir, self.name)
+        cache_dir = os.path.join(cache_dir, self.identifier)
         download_url = _replace_placeholders(
             self.download_url, bundle=bundle.key, version=version, tag=tag
         )
@@ -373,17 +451,17 @@ class Repository(metaclass=ABCMeta):
             progress=progress,
             force=force,
         ) as archive:
-            return bundle.list(archive)
+            return bundle.list(archive, collection)
 
     def clear_cache(self, cache_dir, keep_last_version=True):
-        cache_dir = os.path.join(cache_dir, self.name)
+        cache_dir = os.path.join(cache_dir, self.identifier)
         if not os.path.exists(cache_dir) or not os.path.isdir(cache_dir):
             return
 
         keep = []
         if keep_last_version:
             keep = [
-                bundle.get_filename(tag=DEFAULT_TAG)
+                bundle.get_filename(tag=bundle.tag)
                 for _, bundle in self.get_bundles().items()
             ]
 
@@ -393,12 +471,20 @@ class Repository(metaclass=ABCMeta):
             if (
                 os.path.isfile(full_path)
                 and ext in [".zip", ".sha1"]
-                and not basename in keep
+                and basename not in keep
             ):
                 os.remove(full_path)
 
-    def refresh(self):
+    def refresh(self, timeout=None):
         """Refresh the repository"""
+        try:
+            self._refresh(timeout)
+            self._active = True
+        except requests.ConnectionError:
+            self._active = False
+
+    @abstractmethod
+    def _refresh(self, timeout):
         pass
 
 
@@ -407,8 +493,8 @@ def _validate_url(url):
         return url
     else:
         raise ValueError(
-            "repository url is invalid, got %s ({bundle}, {version} and {tag} are required)"
-            % url
+            "The repository url is invalid, got %s "
+            "({bundle}, {version} and {tag} are required)" % url
         )
 
 
@@ -416,27 +502,63 @@ def _validate_repository_name(str):
     if re.match("[a-zA-Z]+", str):
         return str
     else:
-        raise ValueError("repository name (%s) is not valid" % str)
+        raise ValueError(
+            "The repository name %s is not valid, expected only characters." % str
+        )
 
 
 def _validate_bundle_key(str):
-    if re.match("[a-zA-Z0-9\-]+", str):
+    if re.match(r"[a-zA-Z0-9\-]+", str):
         return str
     else:
-        raise ValueError("bundle key (%s) is not valid" % str)
+        raise ValueError(
+            "The bundle key %s is not valid, expected only characters, numbers or -."
+            % str
+        )
 
 
-def _validate_version(str):
-    if re.match("(^(?:\d+\.)?(?:\d+\.)?(?:\*|\d+)$)", str):
+def _validate_version(str, *, max_version=None):
+    if re.match(r"(^(?:\d+\.)?(?:\d+\.)?(?:\*|\d+)$)", str):
+        if max_version and parse_version(str) > parse_version(max_version):
+            raise ValueError(
+                "The dataset is not supported, required version >=%s, got %s"
+                % (max_version, str)
+            )
         return str
     else:
-        raise ValueError("version (%s) is not valid" % str)
+        raise ValueError("The version %s is not valid." % str)
+
+
+def _validate_collections(collections):
+    if not isinstance(collections, dict):
+        raise TypeError(
+            "'collections' must be dict, not %r" % type(collections).__qualname__
+        )
+    else:
+        for key, values in collections.items():
+            if not isinstance(values, list):
+                raise TypeError(
+                    "'collections.value' must be list, not %r"
+                    % type(values).__qualname__
+                )
+            if not isinstance(key, str):
+                raise TypeError(
+                    "'collections.key' must be str, not" % type(key).__qualname__
+                )
+    return collections
 
 
 class JSONRepository(Repository):
+
+    supported_version = "1.1"
+
     def __init__(self, url):
+        super().__init__()
         self.repo_url = url
-        self.__refresh()
+
+    @property
+    def identifier(self):
+        return base64.b64encode(self.repo_url.encode()).decode("utf-8")
 
     @property
     def wildboar_requires(self):
@@ -457,17 +579,16 @@ class JSONRepository(Repository):
     def get_bundles(self):
         return self._bundles
 
-    def refresh(self):
-        self.__refresh()
-
-    def __refresh(self):
-        json = requests.get(self.repo_url).json()
+    def _refresh(self, timeout):
+        json = requests.get(self.repo_url, timeout=timeout).json()
         self._wildboar_requires = json["wildboar_requires"]
         self._name = _validate_repository_name(json["name"])
-        self._version = _validate_version(json["version"])
+        self._version = _validate_version(
+            json["version"], max_version=JSONRepository.supported_version
+        )
         if parse_version(self.wildboar_requires) > parse_version(wildboar_version):
             raise ValueError(
-                "repository requires wildboar (>=%s), got %s",
+                "The repository requires wildboar >=%s, got %s",
                 self.wildboar_requires,
                 wildboar_version,
             )
@@ -475,16 +596,50 @@ class JSONRepository(Repository):
         bundles = {}
         for bundle_json in json["bundles"]:
             key = _validate_bundle_key(bundle_json["key"])
+            if key in bundles:
+                warnings.warn("duplicate dataset %s (ignoring)" % key, UserWarning)
+
             version = _validate_version(bundle_json["version"])
-            name = bundle_json["name"]
-            class_index = bundle_json["class_index"]
-            description = bundle_json["description"]
-            bundles[bundle_json["key"]] = NpyBundle(
+            tag = bundle_json.get("tag")
+            if tag is not None:
+                tag = _validate_bundle_key(tag)
+
+            name = bundle_json.get("name")
+            if name is None:
+                raise ValueError("bundle %s require 'name', but got None" % key)
+
+            arrays = bundle_json.get("arrays")
+            if arrays is not None:
+                if not isinstance(arrays, list):
+                    warnings.warn(
+                        "'arrays' should be list, not %r. Using defaults."
+                        % type(arrays).__qualname__,
+                        UserWarning,
+                    )
+                    arrays = None
+
+            description = bundle_json.get("description")
+            if description is not None:
+                if not isinstance(description, str):
+                    warnings.warn(
+                        "'description' should be str, not %r. Using defaults."
+                        % type(description).__qualname__,
+                        UserWarning,
+                    )
+                    description = None
+
+            collections = bundle_json.get("collections")
+            if collections is not None:
+                collections = _validate_collections(collections)
+
+            bundles[bundle_json["key"]] = NpBundle(
                 key=key,
                 version=version,
+                tag=tag,
                 name=name,
                 description=description,
-                class_index=class_index,
+                collections=collections,
+                arrays=arrays,
             )
 
         self._bundles = bundles
@@ -492,24 +647,105 @@ class JSONRepository(Repository):
 
 class RepositoryCollection:
     def __init__(self):
-        self.repositories = []
+        self.pending_repositories = set()
+        self.repositories = set()
 
-    def __getitem__(self, item):
-        return next(
-            (repository for repository in self.repositories if repository.name == item),
+    def __getitem__(self, key):
+        repository = next(
+            (repository for repository in self.repositories if repository.name == key),
             None,
         )
 
+        if repository is None:
+            if self.pending_repositories:
+                raise ValueError(
+                    "The repository %s does not exist, but %d repositories have not "
+                    "been refreshed yet." % (key, len(self.pending_repositories))
+                )
+            else:
+                raise ValueError("The repository %s does not exist." % key)
+
+        return repository
+
+    def __delitem__(self, key):
+        self.repositories = set(
+            repository for repository in self.repositories if repository.name != key
+        )
+
     def __contains__(self, item):
-        return any(r for r in self.repositories if r.name == item)
+        return any(
+            repository for repository in self.repositories if repository.name == item
+        )
 
     def __iter__(self):
         return iter(self.repositories)
 
-    def append(self, item):
-        if self[item]:
-            raise ValueError("cannot overwrite repository, %s" % item.name)
-        self.repositories.append(item)
+    def __len__(self):
+        return len(self.repositories)
+
+    def load_repository(self, repository, cache_dir=None):
+        if cache_dir and not repository.active:
+            cache_dir = os.path.join(cache_dir, repository.identifier)
+            if not os.path.exists(cache_dir):
+                os.makedirs(cache_dir, exist_ok=True)
+            repo_path = os.path.join(cache_dir, "repo")
+            if os.path.exists(repo_path):
+                with open(repo_path, "rb") as f:
+                    return pickle.load(f)
+
+        return None
+
+    def save_repository(self, repository, cache_dir=None):
+        if cache_dir and repository.active:
+            cache_dir = os.path.join(cache_dir, repository.identifier)
+            if not os.path.exists(cache_dir):
+                os.makedirs(cache_dir, exist_ok=True)
+            repo_path = os.path.join(cache_dir, "repo")
+            with open(repo_path, "wb") as f:
+                pickle.dump(repository, f)
+
+    def refresh(self, repository=None, timeout=None, cache_dir=None):
+        if repository is None:
+            for repository in self.repositories:
+                repository.refresh(timeout)
+                if repository.active:
+                    self.save_repository(repository, cache_dir=cache_dir)
+
+            for repository in self.pending_repositories:
+                repository.refresh(timeout)
+                if repository.active:
+                    self.repositories.add(repository)
+                    self.save_repository(repository, cache_dir=cache_dir)
+                else:
+                    cached_repository = self.load_repository(
+                        repository, cache_dir=cache_dir
+                    )
+                    if cached_repository:
+                        self.repositories.add(repository)
+
+            self.pending_repositories = set(
+                repository
+                for repository in self.pending_repositories
+                if not repository.active
+            )
+        else:
+            repository = self[repository]
+            repository.refresh(timeout)
+            if repository.active:
+                self.save_repository(repository, cache_dir=cache_dir)
+
+    def install(self, repository, refresh=True, timeout=None, cache_dir=None):
+        if refresh:
+            repository.refresh(timeout)
+
+        if repository.active:
+            self.repositories.add(repository)
+        else:
+            cached_repository = self.load_repository(repository, cache_dir=cache_dir)
+            if cached_repository:
+                self.repositories.add(cached_repository)
+            else:
+                self.pending_repositories.add(repository)
 
 
 class Bundle(metaclass=ABCMeta):
@@ -524,8 +760,9 @@ class Bundle(metaclass=ABCMeta):
     description : str
         Description of the bundle
 
-    class_index : int or array-like
+    label_index : int or array-like
         Index of the class label(s)
+
     """
 
     def __init__(
@@ -534,11 +771,12 @@ class Bundle(metaclass=ABCMeta):
         key,
         version,
         name,
+        tag=None,
+        arrays=None,
         description=None,
-        class_index=-1,
+        collections=None,
     ):
-        """Construct a bundle
-
+        """
         Parameters
         ----------
         key : str
@@ -553,24 +791,35 @@ class Bundle(metaclass=ABCMeta):
         description : str
             Description of the bundle
 
-        class_index : int or array-like
-            Index of the class label(s)
+        arrays : list
+            The arrays of the dataset
         """
         self.key = key
         self.version = version
         self.name = name
         self.description = description
-        self.class_index = class_index
+        self.collections = collections
+        self.tag = tag or DEFAULT_TAG
+        self.arrays = arrays or ["x", "y"]
 
     def get_filename(self, version=None, tag=None, ext=None):
         filename = "%s-v%s" % (self.key, version or self.version)
         if tag:
-            filename += ":%s" % tag
+            filename += "-%s" % tag
         if ext:
             filename += ext
         return filename
 
-    def list(self, archive):
+    def get_collection(self, collection):
+        if self.collections is None:
+            raise ValueError("The collection %s cannot be found." % collection)
+        else:
+            c = self.collections.get(collection)
+            if c is None:
+                raise ValueError("The collection %s cannot be found." % collection)
+            return c
+
+    def list(self, archive, collection=None):
         """List all datasets in this bundle
 
         Parameters
@@ -578,18 +827,26 @@ class Bundle(metaclass=ABCMeta):
         archive : ZipFile
             The bundle file
 
+        collection : str, optional
+            The collection name
+
         Returns
         -------
         dataset_names : list
             A sorted list of datasets in the bundle
         """
         names = []
+        if collection is not None:
+            sample = self.get_collection(collection)
+        else:
+            sample = None
         for f in archive.filelist:
             path, ext = os.path.splitext(f.filename)
             if self._is_dataset(path, ext):
                 filename = os.path.basename(path)
                 filename = re.sub("_(TRAIN|TEST)", "", filename)
-                names.append(filename)
+                if sample is None or filename in sample:
+                    names.append(filename)
 
         return sorted(set(names))
 
@@ -597,8 +854,6 @@ class Bundle(metaclass=ABCMeta):
         self,
         name,
         archive,
-        *,
-        dtype=None,
     ):
         """Load a dataset from the bundle
 
@@ -610,9 +865,6 @@ class Bundle(metaclass=ABCMeta):
         archive : ZipFile
             The zip-file bundle
 
-        dtype : object, optional
-            Cast the data and label matrix to a specific type
-
         Returns
         -------
         x : ndarray
@@ -623,15 +875,18 @@ class Bundle(metaclass=ABCMeta):
 
         n_training_samples : int
             Number of samples that are for training. The value is <= x.shape[0]
+
+        extras : dict, optional
+            Extra numpy arrays
         """
-        dtype = dtype or np.float64
         datasets = []
         for dataset in map(_Dataset, archive.filelist):
             if dataset.filename == name and self._is_dataset(dataset.path, dataset.ext):
                 datasets.append(dataset)
 
         if not datasets:
-            raise ValueError("no dataset found (%s)" % name)
+            raise ValueError("The dataset %s cannot be found." % name)
+
         train_parts = [
             self._load_array(archive, dataset.file)
             for dataset in datasets
@@ -643,22 +898,74 @@ class Bundle(metaclass=ABCMeta):
             if dataset.part == "test"
         ]
 
-        data = np.vstack(train_parts)
-        n_train_samples = data.shape[0]
-        if test_parts:
-            test = np.vstack(test_parts)
-            data = np.vstack([data, test])
+        data = {}
+        for array in self.arrays:
+            data[array] = np.concatenate(
+                [train_part[array] for train_part in train_parts], axis=0
+            )
 
-        y = data[:, self.class_index].astype(dtype)
-        x = np.delete(data, self.class_index, axis=1).astype(dtype)
-        return x, y, n_train_samples
+        sizes = [data[array].shape[0] for array in self.arrays]
+        if max(sizes) != min(sizes):
+            raise ValueError(
+                "Arrays %s must have the same number of samples." % self.arrays
+            )
+
+        n_train_samples = sizes[0]
+        if test_parts:
+            for array in self.arrays:
+                train = data[array]
+                test = []
+                for test_part in test_parts:
+                    test.append(test_part[array])
+                test = np.concatenate(test, axis=0)
+
+                if test.ndim != train.ndim:
+                    raise ValueError(
+                        "The training (%d) and testing (%d) parts have incompatible "
+                        "rank" % (test.ndim, train.ndim)
+                    )
+
+                if test.ndim > 2 and train.shape[1] != test.shape[1]:
+                    raise ValueError(
+                        "The training (%d) and testing (%d) parts have incompatible "
+                        "number dimensions" % (train.shape[1], test.shape[1])
+                    )
+
+                if test.ndim == 1:
+                    data[array] = np.concatenate([train, test], axis=0)
+                else:
+                    if test.ndim == 2:
+                        shape = (
+                            train.shape[0] + test.shape[0],
+                            max(test.shape[-1], train.shape[-1]),
+                        )
+                    elif test.ndim == 3:
+                        shape = (
+                            train.shape[0] + test.shape[0],
+                            test.shape[1],
+                            max(test.shape[-1], train.shape[-1]),
+                        )
+                    else:
+                        raise ValueError(
+                            "Invalid dataset rank %d, expected <= 3" % test.ndim
+                        )
+
+                    merge = np.full(shape, fill_value=EOS, dtype=np.double)
+                    merge[: train.shape[0], ..., : train.shape[-1]] = train
+                    merge[train.shape[0] :, ..., : test.shape[-1]] = test
+                    data[array] = merge
+
+        x = data.pop("x")
+        y = data.pop("y")
+        return x, y, n_train_samples, data
 
     @abstractmethod
     def _is_dataset(self, file_name, ext):
         """Overridden by subclasses
 
-        Check if a path and extension is to be considered a dataset. The check should be simple and only consider
-        the filename and or extension of the file. Validation of the file should be deferred to `_load_array`
+        Check if a path and extension is to be considered a dataset. The check should be
+        simple and only consider the filename and or extension of the file.
+        Validation of the file should be deferred to `_load_array`
 
         Parameters
         ----------
@@ -697,46 +1004,24 @@ class Bundle(metaclass=ABCMeta):
         pass
 
 
-class ArffBundle(Bundle):
-    """bundle of .arff-files"""
-
-    def __init__(
-        self,
-        *,
-        key,
-        version,
-        name,
-        description=None,
-        class_index=-1,
-        encoding="utf-8",
-    ):
-        super().__init__(
-            key=key,
-            version=version,
-            name=name,
-            description=description,
-            class_index=class_index,
-        )
-        self.encoding = encoding
-
-    def _is_dataset(self, file_name, ext):
-        return ext == ".arff"
-
-    def _load_array(self, archive, file):
-        with io.TextIOWrapper(archive.open(file), encoding=self.encoding) as io_wrapper:
-            arff, _metadata = loadarff(io_wrapper)
-            arr = np.array(arff.tolist())
-            return arr
-
-
-class NpyBundle(Bundle):
+class NpBundle(Bundle):
     """bundle of numpy binary files"""
 
     def _is_dataset(self, file_name, ext):
-        return ext == ".npy"
+        return ext in [".npy", ".npz"]
 
     def _load_array(self, archive, file):
-        return np.load(archive.open(file))
+        data = np.load(archive.open(file))
+
+        if isinstance(data, np.ndarray):
+            if "x" in self.arrays and "y" in self.arrays:
+                return {"x": data[:, :-1], "y": data[:, -1]}
+            elif "x" in self.arrays:
+                return {"x": data}
+            else:
+                raise ValueError("Can't infer arrays to export.")
+        else:
+            return data
 
 
 class _Dataset:
