@@ -1,29 +1,15 @@
-# This file is part of wildboar
-#
-# wildboar is free software: you can redistribute it and/or modify it
-# under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# wildboar is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
 # Authors: Isak Samsten
+# License: BSD 3 clause
 
 import numpy as np
 from sklearn.cluster import KMeans
-from sklearn.neighbors import NearestNeighbors, KNeighborsClassifier
+from sklearn.neighbors import KNeighborsClassifier, NearestNeighbors
 from sklearn.utils.validation import check_is_fitted
 
-from .base import BaseCounterfactual
+from ...base import BaseEstimator, CounterfactualMixin, ExplainerMixin
 
 
-class KNeighborsCounterfactual(BaseCounterfactual):
+class KNeighborsCounterfactual(CounterfactualMixin, ExplainerMixin, BaseEstimator):
     """Fit a counterfactual explainer to a k-nearest neighbors classifier
 
     Attributes
@@ -43,17 +29,29 @@ class KNeighborsCounterfactual(BaseCounterfactual):
     def __init__(self, random_state=None):
         self.random_state = random_state
 
-    def fit(self, estimator):
+    def _validate_estimator(self, estimator, allow_3d=False):
         if not isinstance(estimator, KNeighborsClassifier):
             raise ValueError("not a valid estimator")
-        check_is_fitted(estimator)
-        if estimator.metric != "euclidean":
+
+        if estimator.metric == "euclidean" or (
+            estimator.metric == "minkowski" and estimator.p == 2
+        ):
+            return super()._validate_estimator(estimator, allow_3d)
+        else:
             raise ValueError(
-                "only euclidean distance is supported, got %r" % estimator.metric
+                "estimator must be fit with metric='euclidean', got %r"
+                % estimator.metric
             )
 
+    def fit(self, estimator, x=None, y=None):
+        estimator = self._validate_estimator(estimator)
         x = estimator._fit_X
         y = estimator._y
+
+        self.n_timesteps_in_ = estimator.n_features_in_
+        self.n_dims_in_ = 1
+        self.n_features_in_ = self.n_timesteps_in_
+
         classes = estimator.classes_
         n_clusters = x.shape[0] // estimator.n_neighbors
         kmeans = KMeans(n_clusters=n_clusters, random_state=self.random_state).fit(x)
@@ -83,10 +81,10 @@ class KNeighborsCounterfactual(BaseCounterfactual):
         self.explainer_ = label_nn
         return self
 
-    def transform(self, x, y):
-        check_is_fitted(self, ["explainer_"])
-        x_counter = np.empty(x.shape, dtype=x.dtype)
-        success = np.empty(x.shape[0], dtype=bool)
+    def explain(self, x, y):
+        check_is_fitted(self)
+        x, y = self._validate_data(x, y, reset=False, dtype=float)
+        x_counterfactuals = x.copy()
         labels = np.unique(y)
         for label in labels:
             label_indices = np.where(y == label)[0]
@@ -94,8 +92,9 @@ class KNeighborsCounterfactual(BaseCounterfactual):
             nn, mc = self.explainer_[label]
             if mc.shape[0] > 0:
                 closest = nn.kneighbors(x[label_indices, :], return_distance=False)
-                x_counter[label_indices, :] = mc[closest[:, 0], :]
-                success[label_indices] = True
-            else:
-                success[label_indices] = False
-        return x_counter, success
+                x_counterfactuals[label_indices, :] = mc[closest[:, 0], :]
+
+        return x_counterfactuals
+
+    def _more_tags():
+        return {"X_types": []}
