@@ -1,35 +1,29 @@
-# This file is part of wildboar
-#
-# wildboar is free software: you can redistribute it and/or modify it
-# under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# wildboar is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
 # Authors: Isak Samsten
+# License: BSD 3 clause
+
 import abc
 import math
+import numbers
 from copy import deepcopy
 from functools import partial
 
 import numpy as np
 from sklearn.neighbors import NearestNeighbors
-from sklearn.utils.validation import check_is_fitted, check_array, check_random_state
+from sklearn.utils.validation import check_is_fitted, check_random_state, check_scalar
 
-from wildboar.distance.dtw import dtw_mapping, dtw_distance
-from wildboar.distance import distance
-
-from .base import BaseCounterfactual
+from ...base import BaseEstimator, CounterfactualMixin, ExplainerMixin
+from ...distance import pairwise_subsequence_distance
+from ...distance.dtw import (
+    dtw_alignment,
+    dtw_distance,
+    dtw_mapping,
+    wdtw_alignment,
+    wdtw_distance,
+)
+from ...utils.validation import check_option
 
 
-class PrototypeCounterfactual(BaseCounterfactual):
+class PrototypeCounterfactual(CounterfactualMixin, ExplainerMixin, BaseEstimator):
     """Model agnostic approach for constructing counterfactual explanations
 
     Attributes
@@ -56,39 +50,32 @@ class PrototypeCounterfactual(BaseCounterfactual):
 
     def __init__(
         self,
-        background_x,
-        background_y,
-        *,
         metric="euclidean",
-        metric_params=None,
+        *,
+        r=1.0,
+        g=0.05,
         max_iter=100,
         step_size=0.1,
         n_prototypes="auto",
         target="auto",
         method="sample",
-        method_params=None,
-        random_state=None
+        min_shapelet_size=0.0,
+        max_shapelet_size=1.0,
+        random_state=None,
+        verbose=False,
     ):
         """Crate a new model agnostic counterfactual explainer.
 
         Parameters
         ----------
-        background_x : array-like of shape (n_samples, n_timestep)
-            The background data from which prototypes are sampled
-
-        background_y : array-like of shape (n_samples,)
-            The background label from which prototypes are sampled
-
-        metric : {'euclidean', 'dtw'}, optional
+        metric : {'euclidean', 'dtw', 'wdtw'}, optional
             The metric used to move the samples
 
-        metric_params : dict, optional
-            Optional parameters to the metric
+        r : float, optional
+            The warping window size, if metric='dtw' or metric='wdtw'
 
-            If 'dtw':
-
-                r : int or float, optional
-                    The warping window size
+        g : float, optional
+            Penalty control for weighted DTW, if metric='wdtw'
 
         max_iter : int, optional
             The maximum number of iterations
@@ -117,115 +104,169 @@ class PrototypeCounterfactual(BaseCounterfactual):
             - if 'nearest_shapelet' a prototype shapelet is sampled from the
               closest n prototypes
 
-        method_params : dict, optional
-            Additional parameters to the method
+        min_shapelet_size : float, optional
+            Minimum shapelet size, if method='shapelet' or 'nearest_shapelet'
 
-            If 'shapelet' or 'nearest_shapelet'
-
-                min_shapelet_size : float, optional
-                    Minimum shapelet size.
-
-                max_shapelet_size : float, optional
-                    Maximum shapelet size.
+        max_shapelet_size : float, optional
+            Maximum shapelet size, if method='shapelet' or 'nearest_shapelet'
 
         random_state : RandomState or int, optional
             Pseudo-random number for consistency between different runs
         """
-        self.background_x = background_x
-        self.background_y = background_y
         self.random_state = random_state
         self.metric = metric
-        self.metric_params = metric_params
+        self.r = r
+        self.g = g
         self.max_iter = max_iter
         self.step_size = step_size
         self.n_prototypes = n_prototypes
         self.method = method
-        self.method_params = method_params
+        self.min_shapelet_size = min_shapelet_size
+        self.max_shapelet_size = max_shapelet_size
         self.target = target
+        self.verbose = verbose
 
-    def fit(self, estimator):
-        check_is_fitted(estimator)
-        if self.background_x is None or self.background_y is None:
-            raise ValueError("background data are required.")
+    def fit(self, estimator, x, y):
+        if x is None or y is None:
+            raise ValueError("Both training samples and labels are required.")
 
-        x = check_array(self.background_x)
-        y = check_array(self.background_y, ensure_2d=False)
-        if len(y) != x.shape[0]:
-            raise ValueError(
-                "Number of labels={} does not match "
-                "number of samples={}".format(len(y), x.shape[0])
+        estimator = self._validate_estimator(estimator)
+        x, y = self._validate_data(x, y, reset=False, dtype=float)
+
+        metric_params = {}
+        if self.metric in ["dtw", "wdtw"]:
+            metric_params["r"] = check_scalar(
+                self.r, "r", numbers.Real, min_val=0, max_val=1
+            )
+        if self.metric == ["wdtw"]:
+            metric_params["g"] = check_scalar(
+                self.g, "g", numbers.Real, min_val=0, include_boundaries="neither"
             )
-        random_state = check_random_state(self.random_state)
-        metric_params = self.metric_params or {}
-        method_params = self.method_params or {}
-        if self.metric in _METRIC_TRANSFORM:
-            metric = _METRIC_TRANSFORM[self.metric](self.step_size, **metric_params)
-        else:
-            raise ValueError("metric (%s) is not supported" % self.metric)
 
-        if self.method in _PROTOTYPE_SAMPLER:
-            sampler = _PROTOTYPE_SAMPLER[self.method]
-        else:
-            raise ValueError("method (%s) is not supported" % self.method)
+        random_state = check_random_state(self.random_state)
+        metric_transform = check_option(_METRIC_TRANSFORM, self.metric, "metric")(
+            check_scalar(
+                self.step_size,
+                "step_size",
+                numbers.Real,
+                min_val=0,
+                max_val=1,
+                include_boundaries="right",
+            ),
+            **metric_params,
+        )
+        Sampler = check_option(_PROTOTYPE_SAMPLER, self.method, "method")
 
         self.estimator_ = deepcopy(estimator)
-        self.classes_ = np.unique(self.background_y)
+        self.classes_ = np.unique(y)
+        self.random_state_ = random_state.randint(np.iinfo(np.int32).max)
         if self.target == "auto":
             self.target_ = PredictEvaluator(self.estimator_)
         else:
-            if not 0 < self.target <= 1.0:
-                raise ValueError("target must be in (0, 1], got %r" % self.target)
-            self.target_ = ProbabilityEvaluator(self.estimator_, self.target)
+            target = check_scalar(
+                self.target,
+                "target",
+                numbers.Real,
+                min_val=0,
+                max_val=1,
+                include_boundaries="right",
+            )
+            self.target_ = ProbabilityEvaluator(self.estimator_, target)
+
+        if self.method in ["shapelet", "nearest_shapelet"]:
+            min_shapelet_size = check_scalar(
+                self.min_shapelet_size,
+                "min_shapelet_size",
+                numbers.Real,
+                min_val=0,
+                max_val=self.max_shapelet_size,
+            )
+            max_shapelet_size = check_scalar(
+                self.max_shapelet_size,
+                "max_shapelet_size",
+                numbers.Real,
+                min_val=min_shapelet_size,
+                max_val=1.0,
+            )
+            method_params = {
+                "min_shapelet_size": min_shapelet_size,
+                "max_shapelet_size": max_shapelet_size,
+            }
+        else:
+            method_params = {}
 
         self.partitions_ = {}
         for c in self.classes_:
             x_partition = x[y == c]
             if self.n_prototypes == "auto":
                 n_prototypes = x_partition.shape[0]
-            elif isinstance(self.n_prototypes, int):
+            elif isinstance(self.n_prototypes, numbers.Integral):
                 n_prototypes = max(1, min(self.n_prototypes, x_partition.shape[0]))
-            elif isinstance(self.n_prototypes, float):
-                if not 0.0 < self.n_prototypes <= 1.0:
-                    raise ValueError("n_prototypes")
+            elif isinstance(self.n_prototypes, numbers.Real):
+                check_scalar(
+                    self.n_prototypes,
+                    "n_prototypes",
+                    numbers.Real,
+                    min_val=0,
+                    max_val=1,
+                    include_boundaries="right",
+                )
                 n_prototypes = math.ceil(self.n_prototypes * x_partition.shape[0])
             else:
-                raise ValueError("n_prototypes (%r) not supported" % self.n_prototypes)
+                raise ValueError(
+                    "n_prototypes must be 'auto', int or float, got %r"
+                    % self.n_prototypes
+                )
 
-            self.partitions_[c] = sampler(
-                x_partition, c, n_prototypes, metric, random_state, **method_params
+            prototype_indicies = np.arange(x_partition.shape[0])
+            random_state.shuffle(prototype_indicies)
+            prototype_indicies = prototype_indicies[:n_prototypes]
+            self.partitions_[c] = Sampler(
+                x_partition,
+                c,
+                prototype_indicies,
+                metric_transform,
+                **method_params,
             )
 
-    def transform(self, x, y):
-        x = check_array(x)
-        y = check_array(y, ensure_2d=False)
-        if len(y) != x.shape[0]:
-            raise ValueError(
-                "Number of labels={} does not match "
-                "number of samples={}".format(len(y), x.shape[0])
-            )
-        n_samples = x.shape[0]
+    def explain(self, x, y):
+        check_is_fitted(self)
+        random_state = check_random_state(self.random_state_)
+        x, y = self._validate_data(x, y, reset=False, dtype=float)
         counterfactuals = np.empty(x.shape, dtype=x.dtype)
-        success = np.ones(x.shape[0]).astype(bool)
-        for i in range(n_samples):
-            counterfactual = self._transform_sample(x[i], y[i])
-            if counterfactual is not None:
-                counterfactuals[i] = counterfactual
-            else:
-                success[i] = False
-        return counterfactuals, success
+        for i in range(x.shape[0]):
+            if self.verbose:
+                print(f"Computing counterfactual for the {i}th sample.")
 
-    def _transform_sample(self, x, y):
+            counterfactuals[i] = self._transform_sample(
+                x[i], y[i], random_state.randint(np.iinfo(np.int32).max)
+            )
+
+        return counterfactuals
+
+    def _transform_sample(self, x, y, random_state):
+        random_state = check_random_state(random_state)
         sampler = self.partitions_[y]
         o = x.copy()
         n_iter = 0
         while not self.target_.is_counterfactual(o, y) and n_iter < self.max_iter:
-            o = sampler.sample_move(o)
+            if self.verbose and n_iter % (self.max_iter // 10) == 0:
+                print(f"Running {n_iter}/{self.max_iter}...")
+
+            o = sampler.sample_move(o, random_state)
             n_iter += 1
 
-        if n_iter > self.max_iter:
-            return None
-        else:
-            return o
+        if self.verbose:
+            print(f"Completed after {n_iter} iterations.")
+            if n_iter == self.max_iter:
+                print(
+                    "The counterfactual explain reached max_iter, increase step_size "
+                    "or max_iter for convergence."
+                )
+        return o
+
+    def _more_tags():
+        return {"requires_y": True}
 
 
 class TargetEvaluator(abc.ABC):
@@ -246,7 +287,7 @@ class TargetEvaluator(abc.ABC):
 
         Parameters
         ----------
-        x : ndarray of shape (n_samples, n_timestep) or (n_timestep,)
+        x : ndarray of shape (n_timestep,)
             The counterfactual sample
 
         y : object
@@ -267,7 +308,7 @@ class PredictEvaluator(TargetEvaluator):
     """Evaluate if a counterfactual is predicted as y"""
 
     def _is_counterfactual(self, x, y):
-        return self.estimator.predict(x) == y
+        return self.estimator.predict(x)[0] == y
 
 
 class ProbabilityEvaluator(TargetEvaluator):
@@ -298,7 +339,7 @@ class ProbabilityEvaluator(TargetEvaluator):
 
 
 class PrototypeSampler(abc.ABC):
-    def __init__(self, x, y, n_prototypes, metric_transform, random_state):
+    def __init__(self, x, y, prototype_indices, metric_transform):
         """Sample and refine counterfactuals
 
         Parameters
@@ -318,17 +359,12 @@ class PrototypeSampler(abc.ABC):
         random_state : RandomState
             The random number generator.
         """
-        self.random_state = random_state
         self.x = x
         self.y = y
         self.metric_transform = metric_transform
-        self.n_prototypes = n_prototypes
-        prototype_indices = np.arange(x.shape[0])
-        self.random_state.shuffle(prototype_indices)
-        self.prototype_indices = prototype_indices[:n_prototypes]
+        self.prototype_indices = prototype_indices
 
-    @property
-    def _random_index(self):
+    def _get_random_index(self, random_state):
         """Return a random index in the initial prototype sample
 
         Returns
@@ -336,27 +372,11 @@ class PrototypeSampler(abc.ABC):
         int : an index
         """
         return self.prototype_indices[
-            self.random_state.randint(self.prototype_indices.shape[0])
+            random_state.randint(self.prototype_indices.shape[0])
         ]
 
-    def _random_indices(self, n):
-        """Return n random indices from the initial prototype sample
-
-        Parameters
-        ----------
-        n : int
-            The number of idices to return
-
-        Returns
-        -------
-        ndarray : indices
-        """
-        n = min(self.n_prototypes - 1, max(1, n))
-        self.random_state.shuffle(self.prototype_indices)
-        return self.prototype_indices[:n]
-
     @abc.abstractmethod
-    def sample(self, o):
+    def sample(self, o, random_state):
         """Sample an example
 
         Parameters
@@ -389,7 +409,7 @@ class PrototypeSampler(abc.ABC):
         """
         return self.metric_transform.move(o, p)
 
-    def sample_move(self, o):
+    def sample_move(self, o, random_state):
         """Sampla a prototype and move the counterfactual towards the prototype
 
         Parameters
@@ -402,22 +422,23 @@ class PrototypeSampler(abc.ABC):
         new_counterfactual : ndarray of shape (n_timestep,)
             The new counterfactual moved towards the prototype
         """
-        p = self.sample(o)
+        p = self.sample(o, random_state)
         return self.move(o, p)
 
 
 class UniformPrototypeSampler(PrototypeSampler):
     """Sample a prototype uniformly at random from the initial prototype sample"""
 
-    def sample(self, _o):
-        return self.x[self._random_index]
+    def sample(self, _o, random_state):
+        return self.x[self._get_random_index(random_state)]
 
 
 class KNearestPrototypeSampler(PrototypeSampler):
     """Sample a prototype among the samples closest to the current counterfactual"""
 
-    def __init__(self, x, y, n_prototypes, metric_transform, random_state):
-        super().__init__(x, y, n_prototypes, metric_transform, random_state)
+    def __init__(self, x, y, prototype_indicies, metric_transform):
+        super().__init__(x, y, prototype_indicies, metric_transform)
+        n_prototypes = prototype_indicies.size
         if isinstance(self.metric_transform, EuclideanTransform):
             self.nearest_neighbors = NearestNeighbors(
                 metric="euclidean", n_neighbors=n_prototypes
@@ -427,12 +448,19 @@ class KNearestPrototypeSampler(PrototypeSampler):
             self.nearest_neighbors = NearestNeighbors(
                 metric=dtw, n_neighbors=n_prototypes
             )
+        elif isinstance(self.metric_transform, WeightedDynamicTimeWarpTransform):
+            wdtw = partial(
+                wdtw_distance, r=self.metric_transform.r, g=self.metric_transform.g
+            )
+            self.nearest_neighbors = NearestNeighbors(
+                metric=wdtw, n_neighbors=n_prototypes
+            )
         else:
             raise ValueError("unsupported metric")
 
         self.nearest_neighbors.fit(x)
 
-    def nearest_index(self, o):
+    def nearest_index(self, o, random_state):
         """Return the index of the closest sample
 
         Parameters
@@ -447,10 +475,10 @@ class KNearestPrototypeSampler(PrototypeSampler):
         nearest = self.nearest_neighbors.kneighbors(
             o.reshape(1, -1), return_distance=False
         ).reshape(-1)
-        return nearest[self.random_state.randint(nearest.shape[0])]
+        return nearest[random_state.randint(nearest.shape[0])]
 
-    def sample(self, o):
-        return self.x[self.nearest_index(o)]
+    def sample(self, o, random_state):
+        return self.x[self.nearest_index(o, random_state)]
 
 
 class ShapeletPrototypeSampler(PrototypeSampler):
@@ -460,9 +488,8 @@ class ShapeletPrototypeSampler(PrototypeSampler):
         self,
         x,
         y,
-        n_prototypes,
+        prototype_indicies,
         metric_transform,
-        random_state,
         min_shapelet_size=0,
         max_shapelet_size=1,
     ):
@@ -488,11 +515,11 @@ class ShapeletPrototypeSampler(PrototypeSampler):
         max_shapelet_size : float
             The maximum shapelet size
         """
-        super().__init__(x, y, n_prototypes, metric_transform, random_state)
+        super().__init__(x, y, prototype_indicies, metric_transform)
         self.min_shapelet_size = min_shapelet_size
         self.max_shapelet_size = max_shapelet_size
 
-    def sample_shapelet(self, p):
+    def sample_shapelet(self, p, random_state):
         """Sample a shapelet from x
 
         Parameters
@@ -506,16 +533,18 @@ class ShapeletPrototypeSampler(PrototypeSampler):
             A shapelet
         """
         n_timestep = self.x.shape[-1]
-        min_shapelet_size = max(2, int(n_timestep * self.min_shapelet_size))
-        max_shapelet_size = int(n_timestep * self.max_shapelet_size)
-        shapelet_length = self.random_state.randint(
-            min_shapelet_size, max_shapelet_size
+        min_shapelet_size = max(
+            min(2, n_timestep), math.ceil(n_timestep * self.min_shapelet_size)
         )
-        start_index = self.random_state.randint(0, n_timestep - shapelet_length)
+        max_shapelet_size = math.ceil(n_timestep * self.max_shapelet_size)
+        shapelet_length = random_state.randint(min_shapelet_size, max_shapelet_size)
+        start_index = random_state.randint(0, n_timestep - shapelet_length)
         return p[start_index : (start_index + shapelet_length)]
 
-    def sample(self, _o):
-        return self.sample_shapelet(self.x[self._random_index])
+    def sample(self, _o, random_state):
+        return self.sample_shapelet(
+            self.x[self._get_random_index(random_state)], random_state
+        )
 
     def move(self, o, p):
         """Move the best matching shapelet of the  counterfactual sample towards
@@ -540,12 +569,19 @@ class ShapeletPrototypeSampler(PrototypeSampler):
         elif isinstance(self.metric_transform, DynamicTimeWarpTransform):
             metric = "dtw"
             metric_params = {"r": self.metric_transform.r}
+        elif isinstance(self.metric_transform, WeightedDynamicTimeWarpTransform):
+            metric = "wdtw"
+            metric_params = {"r": self.metric_transform.r, "g": self.metric_transform.g}
         else:
             raise ValueError("unsupported metric")
 
         # Find the best matching position in
-        min_dist, best_match = distance(
-            p, o, metric=metric, metric_params=metric_params, return_index=True
+        min_dist, best_match = pairwise_subsequence_distance(
+            p,
+            o,
+            metric=metric,
+            metric_params=metric_params,
+            return_index=True,
         )
         o[best_match : best_match + p.shape[0]] = self.metric_transform.move(
             o[best_match : best_match + p.shape[0]], p
@@ -564,29 +600,27 @@ class KNearestShapeletPrototypeSampler(PrototypeSampler):
         self,
         x,
         y,
-        n_prototypes,
+        prototype_indicies,
         metric_transform,
-        random_state,
         min_shapelet_size=0,
         max_shapelet_size=1,
     ):
-        super().__init__(x, y, n_prototypes, metric_transform, random_state)
+        super().__init__(x, y, prototype_indicies, metric_transform)
         self.nearest_sampler = KNearestPrototypeSampler(
-            x, y, n_prototypes, metric_transform, random_state
+            x, y, prototype_indicies, metric_transform
         )
         self.shapelet_sampler = ShapeletPrototypeSampler(
             x,
             y,
-            n_prototypes,
+            prototype_indicies,
             metric_transform,
-            random_state,
             min_shapelet_size,
             max_shapelet_size,
         )
 
-    def sample(self, o):
-        p = self.nearest_sampler.nearest_index(o)
-        return self.shapelet_sampler.sample_shapelet(self.x[p])
+    def sample(self, o, random_state):
+        p = self.nearest_sampler.nearest_index(o, random_state)
+        return self.shapelet_sampler.sample_shapelet(self.x[p], random_state)
 
     def move(self, o, p):
         return self.shapelet_sampler.move(o, p)
@@ -605,8 +639,6 @@ class MetricTransform(abc.ABC):
             sample is moved less and values closer to 1 mean that the sample
             is moved more.
         """
-        if not 0.0 < gamma <= 1.0:
-            raise ValueError("gamma must be in (0, 1], got %r" % gamma)
         self.gamma = gamma
 
     @abc.abstractmethod
@@ -616,13 +648,14 @@ class MetricTransform(abc.ABC):
         Parameters
         ----------
         o : ndarray of shape (n_timestep,)
-            An array
+            An array, the time series to move
+
         p : ndarray of shape (n_timestep,)
-            An array
+            An array, the time series to move towards
 
         Returns
         -------
-        ndarray : an array
+        ndarray : an array, the result of moving o closer to p
         """
         pass
 
@@ -642,18 +675,36 @@ class DynamicTimeWarpTransform(MetricTransform):
         self.r = r
 
     def move(self, o, p):
-        indicator, indices = dtw_mapping(o, p, r=self.r, return_index=True)
+        _, indices = dtw_mapping(alignment=self._get_alignment(o, p), return_index=True)
+
+        # This a fast but somewhat crude approximation of DBA (as implemented) in
+        # dtw_average with sample_weight=[(1 - gamma), gamma] and init=0 but
+        # significantly faster to compute.
         result = o * (1 - self.gamma)
         weight = np.ones(o.shape[0]) * (1 - self.gamma)
-        for i, j in zip(*indices):
-            result[j] += self.gamma * p[j]
-            weight[j] += self.gamma
+        for i, j in zip(indices[0], indices[1]):
+            result[i] += self.gamma * p[j]
+            weight[i] += self.gamma
+
         return result / weight
 
+    def _get_alignment(self, o, p):
+        return dtw_alignment(o, p, r=self.r)
+
+
+class WeightedDynamicTimeWarpTransform(DynamicTimeWarpTransform):
+    def __init__(self, gamma, r=1, g=0.05):
+        super().__init__(gamma, r)
+        self.g = g
+
+    def _get_alignment(self, o, p):
+        return wdtw_alignment(o, p, r=self.r, g=self.g)
+
 
 _METRIC_TRANSFORM = {
     "euclidean": EuclideanTransform,
     "dtw": DynamicTimeWarpTransform,
+    "wdtw": WeightedDynamicTimeWarpTransform,
 }
 
 _PROTOTYPE_SAMPLER = {
